"""
Bonito Aligner
"""
from threading import Thread
from functools import partial
from mappy import Aligner, ThreadBuffer

from bonito.multiprocessing import ThreadMap


def align_map(aligner, sequences, n_thread=4):
    """
    Align `sequences` with minimap using `n_thread` threads.
    """
    return ThreadMap(partial(MappyWorker, aligner), sequences, n_thread)


class ManagedThreadBuffer:
    """
    Minimap2 ThreadBuffer that is periodically reallocated.
    """
    def __init__(self, max_uses=20):
        self.max_uses = max_uses
        self.uses = 0
        self._b = ThreadBuffer()

    @property
    def buffer(self):
        if self.uses > self.max_uses:
            self._b = ThreadBuffer()
            self.uses = 0
        self.uses += 1
        return self._b


class MappyWorker(Thread):
    """
    Process that reads items from an input_queue, applies a func to them and puts them on an output_queue
    """
    def __init__(self, aligner, input_queue=None, output_queue=None):
        super().__init__()
        self.aligner = aligner
        self.input_queue = input_queue
        self.output_queue = output_queue

    def run(self):
        thrbuf = ManagedThreadBuffer()
        while True:
            item = self.input_queue.get()
            if item is StopIteration:
                self.output_queue.put(item)
                break
            k, v = item
            mapping = next(self.aligner.map(v['sequence'], buf=thrbuf.buffer, MD=True), None)
            self.output_queue.put((k, {**v, 'mapping': mapping}))



#!/usr/bin/env python3

"""
Bonito training.
"""

import os
from pathlib import Path
from importlib import import_module
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser

import toml
import torch

from bonito.training import Trainer
from bonito.data import load_data, ModelSetup, ComputeSettings, DataSettings
from bonito.util import __models_dir__, default_config, load_model, load_symbol, init


def main(args):
    workdir = os.path.expanduser(args.training_directory)
    if os.path.exists(workdir) and not args.force:
        print("[error] %s exists, use -f to force continue training." % workdir)
        exit(1)
    os.makedirs(workdir, exist_ok=True)

    init(args.seed, args.device, (not args.nondeterministic))
    device = torch.device(args.device)

    if not args.pretrained:
        config = toml.load(args.config)
    else:
        dirname = args.pretrained
        if not os.path.isdir(dirname) and os.path.isdir(os.path.join(__models_dir__, dirname)):
            dirname = os.path.join(__models_dir__, dirname)
        pretrain_file = os.path.join(dirname, 'config.toml')
        config = toml.load(pretrain_file)
        if 'lr_scheduler' in config:
            print("[ignoring 'lr_scheduler' in --pretrained config]")
            del config['lr_scheduler']

    # Record the args and working directory so we can track them
    argsdict = dict(training=vars(args))
    argsdict["training"]["pwd"] = os.getcwd()

    print("[loading model]")
    if args.pretrained:
        print(f"[using pretrained model {args.pretrained}]")
        model = load_model(args.pretrained, device, half=False)
    else:
        model = load_symbol(config, 'Model')(config)

    print("[loading data]")
    data = DataSettings(
        training_data=args.directory,
        num_train_chunks=args.chunks,
        num_valid_chunks=args.valid_chunks,
        output_dir=workdir
    )
    model_setup = ModelSetup(
        n_pre_context_bases=getattr(model, "n_pre_context_bases", 0),
        n_post_context_bases=getattr(model, "n_post_context_bases", 0),
        standardisation=config.get("standardisation", {}),
    )
    compute_settings = ComputeSettings(
        batch_size=args.batch,
        num_workers=args.num_workers,
        seed=args.seed,
    )

    train_loader, valid_loader = load_data(data, model_setup, compute_settings)

    try:
        # Allow the train-loader to write meta-data fields to the config
        dataset_cfg = train_loader.dataset.dataset_config
    except AttributeError:
        dataset_cfg = {}
    toml.dump({**config, **argsdict, **dataset_cfg}, open(os.path.join(workdir, 'config.toml'), 'w'))

    if config.get("lr_scheduler"):
        sched_config = config["lr_scheduler"]
        lr_scheduler_fn = getattr(
            import_module(sched_config["package"]), sched_config["symbol"]
        )(**sched_config)
    else:
        lr_scheduler_fn = None

    trainer = Trainer(
        model, device, train_loader, valid_loader,
        use_amp=not args.no_amp,
        lr_scheduler_fn=lr_scheduler_fn,
        restore_optim=args.restore_optim,
        save_optim_every=args.save_optim_every,
        grad_accum_split=args.grad_accum_split,
        quantile_grad_clip=args.quantile_grad_clip,
        chunks_per_epoch=args.chunks,
        batch_size=args.batch,
    )

    if (',' in args.lr):
        lr = [float(x) for x in args.lr.split(',')]
    else:
        lr = float(args.lr)
    optim_kwargs = config.get("optim", {})
    trainer.fit(workdir, args.epochs, lr, **optim_kwargs)

def argparser():
    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        add_help=False
    )
    parser.add_argument("training_directory")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--config', default=default_config)
    group.add_argument('--pretrained', default="")
    parser.add_argument("--directory", type=Path)
    parser.add_argument("--device", default="cuda")
    parser.add_argument("--lr", default='2e-3')
    parser.add_argument("--seed", default=25, type=int)
    parser.add_argument("--epochs", default=5, type=int)
    parser.add_argument("--batch", default=64, type=int)
    parser.add_argument("--chunks", default=0, type=int)
    parser.add_argument("--valid-chunks", default=None, type=int)
    parser.add_argument("--no-amp", action="store_true", default=False)
    parser.add_argument("-f", "--force", action="store_true", default=False)
    parser.add_argument("--restore-optim", action="store_true", default=False)
    parser.add_argument("--nondeterministic", action="store_true", default=False)
    parser.add_argument("--save-optim-every", default=10, type=int)
    parser.add_argument("--grad-accum-split", default=1, type=int)
    quantile_group = parser.add_mutually_exclusive_group()
    quantile_group.add_argument('--quantile-grad-clip', dest='quantile_grad_clip', action='store_true')
    quantile_group.add_argument('--no-quantile-grad-clip', dest='quantile_grad_clip', action='store_false')
    quantile_group.set_defaults(quantile_grad_clip=True)
    parser.add_argument("--num-workers", default=4, type=int)
    return parser




import torch
from crf_beam import beam_search
from collections import namedtuple


def max_grad(x, dim=0):
    return torch.zeros_like(x).scatter_(dim, x.argmax(dim, True), 1.0)


semiring = namedtuple('semiring', ('zero', 'one', 'mul', 'sum', 'dsum'))
Log = semiring(zero=-1e38, one=0., mul=torch.add, sum=torch.logsumexp, dsum=torch.softmax)
Max = semiring(zero=-1e38, one=0., mul=torch.add, sum=(lambda x, dim=0: torch.max(x, dim=dim)[0]), dsum=max_grad)


def scan(Ms, idx, v0, S:semiring=Log):
    T, N, C, NZ = Ms.shape
    alpha = Ms.new_full((T + 1, N, C), S.zero)
    alpha[0] = v0
    for t in range(T):
        alpha[t+1] = S.sum(S.mul(Ms[t], alpha[t, :, idx]), dim=-1)
    return alpha


class CTC_CRF_Torch:

    def __init__(self, state_len=5, alphabet='NACGT'):
        super().__init__()
        self.alphabet = alphabet
        self.state_len = state_len
        self.n_base = len(alphabet[1:])
        self.idx = torch.cat([
            torch.arange(self.n_base**(self.state_len))[:, None],
            torch.arange(
                self.n_base**(self.state_len)
            ).repeat_interleave(self.n_base).reshape(self.n_base, -1).T
        ], dim=1).to(torch.int32)

    def forward_scores(self, scores, S: semiring=Log):
        T, N, C = scores.shape
        Ms = scores.reshape(T, N, -1, self.n_base + 1)
        v0 = Ms.new_full((N, self.n_base**(self.state_len)), S.one)
        return scan(Ms, self.idx.to(torch.int64), v0, S)

    def backward_scores(self, scores, S: semiring=Log):
        T, N, _ = scores.shape
        vT = scores.new_full((N, self.n_base**(self.state_len)), S.one)
        idx_T = self.idx.flatten().argsort().reshape(*self.idx.shape)
        Ms_T = scores[:, :, idx_T]
        idx_T = torch.div(idx_T, self.n_base + 1, rounding_mode='floor')
        return scan(Ms_T.flip(0), idx_T.to(torch.int64), vT, S).flip(0)
    
    def posteriors(self, scores, S: semiring=Log):
        fwd = model.forward_scores(scores, S)
        bwd = model.backward_scores(scores, S)
        return torch.softmax(fwd + bwd, dim=-1)

    def viterbi(self, scores):
        traceback = self.posteriors(scores, Max)
        paths = traceback.argmax(2) % len(self.alphabet)
        return paths


def compute_scores_torch(scores):
    model = CTC_CRF_Torch()
    fwd = model.forward_scores(scores)
    bwd = model.backward_scores(scores)
    posts = torch.softmax(fwd + bwd, dim=-1)

    return {
        'scores': scores.transpose(0, 1).cpu(),
        'bwd': bwd.transpose(0, 1).cpu(),
        'posts': posts.transpose(0, 1).cpu(),
    }


def decode_cpu(x, beam_width=32, beam_cut=100.0, scale=1.0, offset=0.0, blank_score=2.0, rna=False):
    sequence, qstring, moves = beam_search(x['scores'], x['bwd'], x['posts'], beam_size=beam_width)
    return {
        'sequence': sequence,
        'qstring': qstring,
        'moves': moves,
    }



import subprocess
import unittest

from bonito import modules

class TestCli(unittest.TestCase):
    """
    Minimal test case that each of the CLI entry points can be called from terminal
    """

    def test_tool_gets_help(self):
        for tool in modules:
            help_message = subprocess.check_output(["bonito", tool, "-h"])
            self.assertTrue(f"usage: bonito {tool}".encode() in help_message)



import subprocess
import unittest
from io import StringIO
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock

from bonito.cli.download import models, training, Downloader


@mock.patch("sys.stdout", new=StringIO())
@mock.patch("sys.stderr", new=StringIO())
class TestDownload(unittest.TestCase):

    def setUp(self):
        self._tmp_dir = TemporaryDirectory()
        self.save_path = Path(self._tmp_dir.name)

    def tearDown(self):
        self._tmp_dir.cleanup()

    def test_list(self):
        bonito_args = ["bonito", "download", "--list"]
        proc = subprocess.Popen(bonito_args + ["--models"], stderr=subprocess.PIPE)
        _, stderr = proc.communicate()
        self.assertTrue(len(stderr.decode().split('\n')) - 1, len(models))

        proc = subprocess.Popen(bonito_args + ["--training"], stderr=subprocess.PIPE)
        _, stderr = proc.communicate()
        self.assertTrue(len(stderr.decode().split('\n')) - 1, len(training))

        proc = subprocess.Popen(bonito_args + ["--all"], stderr=subprocess.PIPE)
        _, stderr = proc.communicate()
        self.assertTrue(len(stderr.decode().split('\n')) - 1, len(models) + len(training))

    def test_download(self):
        downloader = Downloader(out_dir=self.save_path)
        model_subset = models[:3]  # Pick the most recent models as an example
        for model in model_subset:
            downloader.download(model)
            self.assertTrue((self.save_path / model).exists())

        # Check we've cleaned up any other files
        assert len(list(self.save_path.iterdir())) == len(model_subset)

    def test_force(self):
        downloader = Downloader(out_dir=self.save_path, force=False)
        mock_out = self.save_path / "mock_output"
        mock_out.write_text("placeholder_text")

        # This should run without error despite the url not existing
        _ = downloader.download(mock_out.name)

        # This should fail to download from the CDN
        downloader.force = True
        with self.assertRaises(FileNotFoundError):
            _ = downloader.download(mock_out.name)



"""
Bonito Export
"""
import logging
import shutil
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from pathlib import Path

import toml
import torch

from bonito.nn import fuse_bn_, Clamp
from bonito.util import _load_model, get_last_checkpoint

logger = logging.getLogger(__name__)


def save_tensor(directory, name, tensor):
    """
    Save a tensor `x` to `fn.tensor` for use with libtorch.
    """
    module = torch.nn.Module()
    param = torch.nn.Parameter(tensor, requires_grad=False)
    module.register_parameter("0", param)
    tensors = torch.jit.script(module)
    tensors.save(directory / f"{name}.tensor")


def clean_config(config):
    """
    Strip any non-inference time features out of the model
    """
    config.pop("decoder", None)
    config.pop("aux_CRF_losses", None)
    config.pop("training", None)
    config.pop("basecaller", None)
    config.pop("lr_scheduler", None)
    config.pop("optim", None)

    expected_fields = ["qscore", "run_info", "scaling", "standardisation", "training_dataset"]
    for field in expected_fields:
        if field not in config:
            logger.warning(f"INFO: metadata '{field}' is not set in config")
    return config


def get_layer_order_map(base_encoder):
    # For models with clamp layers we have to reorder the output layers
    # so that Dorado can parse them correctly
    clamp_count = 0
    layer_order_map = {}
    for i, layer in enumerate(base_encoder):
        if isinstance(layer, Clamp):
            clamp_count += 1
        layer_order_map[str(i)] = str(i - clamp_count)
    return layer_order_map


def export_to_dorado(model, config_dict, output):
    output.mkdir(exist_ok=True, parents=True)

    if hasattr(model, "base_model") and hasattr(model.base_model, "encoder"):
        # v5-style transformer models
        encoder = model.base_model.encoder
        config_dict["model"] = config_dict["model"]["base_model"]
    elif hasattr(model, "encoder") and hasattr(model.encoder, "base_encoder"):
        # v4-style lstm-models
        encoder = model.encoder.base_encoder
    else:
        encoder = model.encoder

    config_dict = clean_config(config_dict)
    with (output / "config.toml").open("w") as f:
        toml.dump(config_dict, f)

    for name, tensor in encoder.state_dict().items():
        save_tensor(output, name, tensor)

    if any(isinstance(encoder[i], Clamp) for i in range(len(encoder) - 1)):
        reorder_layers_without_clamp(encoder, output)


def reorder_layers_without_clamp(encoder, output):
    # In v4.0-v4.2 we had clamp layers after the convs which need removing
    layer_order_map = get_layer_order_map(encoder)
    for name, tensor in encoder.state_dict().items():
        # Rename the layers to avoid counting Clamps
        # We have to do this _after_ saving the file to get an identical object
        # since tensor.save() encodes the filename in the file
        old_layer_id = name.split(".")[0]
        new_layer_id = layer_order_map.get(old_layer_id, old_layer_id)
        new_name = name.replace(old_layer_id, new_layer_id, 1)
        if name != new_name:
            shutil.move(output / f"{name}.tensor", output / f"{new_name}.tensor")


def main(args):
    export_model(args.model, args.output, args.config)


def export_model(model, output, config_file):
    if config_file is None:
        config_file = model / "config.toml"

    config_dict = toml.load(config_file)
    model_file = get_last_checkpoint(model) if model.is_dir() else model
    model = _load_model(model_file, config_dict, device='cpu')

    # fuse conv+batchnorm
    # model weights might be saved in half when training and PyTorch's bn fusion
    # code uses an op (rsqrt) that currently (1.11) only has a float implementation
    model = model.to(torch.float32).apply(fuse_bn_)

    export_to_dorado(model, config_dict, output)


def argparser():
    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        add_help=False
    )
    parser.add_argument('model', type=Path)
    parser.add_argument('--output', type=Path, required=True)
    parser.add_argument('--config', type=Path, default=None,
                        help='config file to read settings from')
    return parser



"""
Bonito Read Utils
"""
import sys
from glob import iglob
from collections import OrderedDict
from importlib import import_module

import torch
import numpy as np


__formats__ = ["pod5", "fast5"]

# Normalisation parameters for kit 14 DNA
# Different parameters can be specified in the 'normalisation' section
# of a bonito config file
__default_norm_params__ = {'quantile_a' : 0.2,
                           'quantile_b' : 0.9,
                           'shift_multiplier' : 0.51,
                           'scale_multiplier' : 0.53}

class Reader:

    def __init__(self, directory, recursive=False):
        self.fmt = None
        for fmt in __formats__:
            pattern = f"**/*.{fmt}" if recursive else f"*.{fmt}"
            match = next(iglob(directory + "/" + pattern, recursive=True), None)
            if match is not None:
                self.fmt = fmt
                break
        else:
            raise FileNotFoundError()
        if self.fmt == "fast5":
            sys.stderr.write(
                "DeprecationWarning: fast5 support will be deprecated in a "
                "future bonito version. Please use pod5\n"
            )
        _reader = import_module(f"bonito.{self.fmt}")
        self._get_reads = getattr(_reader, "get_reads")
        self._get_read_groups = getattr(_reader, "get_read_groups")

    def get_reads(self, *args, **kwargs):
        return self._get_reads(*args, **kwargs)

    def get_read_groups(self, *args, **kwargs):
        return self._get_read_groups(*args, **kwargs)


class Read:

    def __init__(self, read, filename, meta=False):
        ...

    def __repr__(self):
        return "Read('%s')" % self.read_id

    def readgroup(self, model):
        self._groupdict = OrderedDict([
            ('ID', f"{self.run_id}_{model}"),
            ('PL', f"ONT"),
            ('DT', f"{self.exp_start_time}"),
            ('PU', f"{self.flow_cell_id}"),
            ('PM', f"{self.device_id}"),
            ('LB', f"{self.sample_id}"),
            ('SM', f"{self.sample_id}"),
            ('DS', f"%s" % ' '.join([
                f"run_id={self.run_id}",
                f"basecall_model={model}",
            ]))
        ])
        return '\t'.join(["@RG", *[f"{k}:{v}" for k, v in self._groupdict.items()]])

    def tagdata(self):
        return [
            f"mx:i:{self.mux}",
            f"ch:i:{self.channel}",
            f"st:Z:{self.start_time}",
            f"du:f:{self.duration}",
            f"rn:i:{self.read_number}",
            f"f5:Z:{self.filename}",
            f"sm:f:{self.shift}",
            f"sd:f:{self.scale}",
            f"sv:Z:{self.scaling_strategy}",
        ]


class ReadChunk:

    def __init__(self, read, chunk, i, n):
        self.read_id = "%s:%i:%i" % (read.read_id, i, n)
        self.run_id = read.run_id
        self.filename = read.filename
        self.mux = read.mux
        self.channel = read.channel
        self.start = read.start
        self.duration = read.duration
        self.template_start = self.start
        self.template_duration = self.duration
        self.signal = chunk

    def __repr__(self):
        return "ReadChunk('%s')" % self.read_id


def read_chunks(read, chunksize=4000, overlap=400):
    """
    Split a Read in fixed sized ReadChunks
    """
    if len(read.signal) < chunksize:
        return

    _, offset = divmod(len(read.signal) - chunksize, chunksize - overlap)
    signal = torch.from_numpy(read.signal[offset:])
    blocks = signal.unfold(0, chunksize, chunksize - overlap)

    for i, block in enumerate(blocks):
        yield ReadChunk(read, block.numpy(), i+1, blocks.shape[0])


def trim(signal, window_size=40, threshold=2.4, min_trim=10, min_elements=3, max_samples=8000, max_trim=0.3):

    seen_peak = False
    num_windows = min(max_samples, len(signal)) // window_size

    for pos in range(num_windows):
        start = pos * window_size + min_trim
        end = start + window_size
        window = signal[start:end]
        if len(window[window > threshold]) > min_elements or seen_peak:
            seen_peak = True
            if window[-1] > threshold:
                continue
            if end >= min(max_samples, len(signal)) or end / len(signal) > max_trim:
                return min_trim
            return end

    return min_trim


def normalisation(sig, scaling_strategy=None, norm_params=None):
    """
    Calculate signal shift and scale factors for normalisation or standardisation.
    If no information is provided in the config, quantile scaling is default.
    """
    if scaling_strategy and scaling_strategy.get("strategy") == "pa":
        if norm_params and norm_params.get("standardise") == 1:
            shift = norm_params.get('mean')
            scale = norm_params.get('stdev')
        elif norm_params and norm_params.get("standardise") == 0:
            shift = 0.0
            scale = 1.0
        else:
            raise ValueError("Picoampere scaling requested, but standardisation flag not provided")

    elif scaling_strategy is None or scaling_strategy.get("strategy") == "quantile":
        if norm_params is None:
            norm_params = __default_norm_params__

        qa, qb = np.quantile(sig, [norm_params['quantile_a'], norm_params['quantile_b']])
        shift = max(10, norm_params['shift_multiplier'] * (qa + qb))
        scale = max(1.0, norm_params['scale_multiplier'] * (qb - qa))
    else:
        raise ValueError(f"Scaling strategy {scaling_strategy.get('strategy')} not supported; choose quantile or pa.")
    return shift, scale
